import logging
import sys
import time
import json
from typing import Any, Callable, Dict, List, Optional, Union, Tuple
from threading import Lock, Thread
from multiprocessing.pool import ThreadPool
import grpc
from formant.protos.agent.v1 import agent_pb2, agent_pb2_grpc
from formant.protos.model.v1 import (
    commands_pb2,
    datapoint_pb2,
    event_pb2,
    math_pb2,
    media_pb2,
    navigation_pb2,
    health_pb2,
    text_pb2,
    intervention_pb2,
    file_pb2,
)
from typing_extensions import Literal
from grpc_status import rpc_status


from formant.sdk.agent.v1.localization.localization_manager import LocalizationManager

from .exceptions import (
    InvalidArgument,
    handle_agent_exceptions,
    handle_grpc_exceptions,
    handle_post_data_exceptions
    )
from .cancellable_stream_thread import CancellableStreamThread

import os

DEFAULT_AGENT_URL = "unix:///var/lib/formant/agent.sock"
DEFAULT_IMAGE_CONTENT_TYPE = "image/jpg"  # type: Literal["image/jpg"]
DEFAULT_SEVERITY_TYPE = "info"  # type: Literal["info"]
ALLOWED_IMAGE_CONTENT_TYPES = ["image/jpg", "image/jpeg", "image/png"]
ALLOWED_VIDEO_CONTENT_TYPES = ["video/h264"]
ALLOWED_IMAGE_AND_VIDEO_CONTENT_TYPES = ALLOWED_IMAGE_CONTENT_TYPES + ALLOWED_VIDEO_CONTENT_TYPES
ALLOWED_SEVERITY_TYPES = ["info", "warning", "critical", "error"]
DEFAULT_LOCALIZATION_STEAM_NAME = "localization"
DEFAULT_THROTTLE_HZ = 5


def current_timestamp():
    return int(time.time() * 1000)


def set_severity_pb(
    given,  # type: str
):
    severity_pb = event_pb2.INFO
    if given == "warning":
        severity_pb = event_pb2.WARNING
    elif given == "critical":
        severity_pb = event_pb2.CRITICAL
    elif given == "error":
        severity_pb = event_pb2.ERROR
    return severity_pb


def validate_string_in_array(
    given,  # type: Any
    expected_types,  # type: List[str]
    input_name,  # type: str
):
    method_name = sys._getframe(1).f_code.co_name
    if given not in expected_types:
        raise TypeError(
            method_name
            + input_name
            + " is "
            + given
            + "' but must be one of: "
            + str(expected_types)
        )


def validate_type(
    given,  # type: Any
    expected_types,  # type: List[type]
):
    if not any([isinstance(given, t) for t in expected_types]):
        method_name = sys._getframe(1).f_code.co_name

        # e.g.
        # post_text input 'None' has type <class 'NoneType'>,
        # but expected one of: [<class 'str'>]
        raise TypeError(
            method_name
            + " input '"
            + str(given)
            + "' has type "
            + str(type(given))
            + ", but expected one of: "
            + str(expected_types)
        )


class Client:
    """
    A client for interacting with the Formant agent.
    Automatically handles connection and reconnection to the agent.
    There are methods for:

    * Ingesting telemetry datapoints

    * Creating events

    * Handling commands

    * Ingesting transform frames

    * Reading application configuration

    * Handling teleop control datapoints
    """

    def __init__(
        self,
        agent_url=DEFAULT_AGENT_URL,  # type: str
        enable_logging=True,  # type: bool
        ignore_throttled=False,  # type: bool
        ignore_unavailable=False,  # type: bool
        local_dev=False,  # type: bool
        thread_pool_size=10,  # type: int
    ):
        """
        :param agent_url: The address of the Formant agent API.
        :param enable_logging: If ``True``, this client will log some information to ``stdout``.
        :param ignore_throttled: If ``True``, ``telemetry datapoint throttle`` errors
            will not raise Exceptions. Throttled datapoints are still valid for teleoperation.
        :param ignore_unavailable: If ``True``, ``Formant agent unavailable`` errors
            will not raise Exceptions.
        """

        self.logger = logging.getLogger("formant")  # type: Optional[logging.Logger]

        if enable_logging:
            self.logger.setLevel(logging.INFO)
            handler = logging.StreamHandler(sys.stdout)
            formatter = logging.Formatter("%(levelname)s: %(message)s")
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)
        else:
            self.logger.addHandler(logging.NullHandler())

        self.ignore_throttled = ignore_throttled
        self.ignore_unavailable = ignore_unavailable
        self._lock = Lock()
        self._thread_pool = ThreadPool(processes=thread_pool_size)
        self._agent_url = agent_url  # type: str
        if local_dev == True:
            self._agent_url = os.getenv(
                "FORMANT_AGENT_URL", DEFAULT_AGENT_URL
            )  # type: str
        self._app_config = {}  # type: Dict[str, str]
        self._config_update_callbacks = []  # type: List[Callable[[], None]]
        self._command_request_callback_streams = (
            {}
        )  # type: Dict[Callable[..., None], CancellableStreamThread]
        self._teleop_callback_streams = (
            {}
        )  # type: Dict[Callable[..., None], CancellableStreamThread]
        self._teleop_heartbeat_callback_streams = (
            {}
        )  # type: Dict[Callable[..., None], CancellableStreamThread]
        self._telemetry_listener_callback_streams = (
            {}
        )  # type: Dict[Callable[..., None], CancellableStreamThread]
        self._custom_data_channel_message_callback_streams = (
            {}
        )  # type: Dict[Callable[..., None], CancellableStreamThread]
        self._localization_manager_cache = {}  # type: Dict[str,LocalizationManager]
        self._connected = False
        self._setup_agent_communication()

    def get_localization_manager(
        self,
        stream_name=DEFAULT_LOCALIZATION_STEAM_NAME,  # type: str
        throttle_hz=DEFAULT_THROTTLE_HZ,  # type:float
    ):
        # type: (...) -> LocalizationManager
        validate_type(stream_name, [str])
        if self._localization_manager_cache.get(stream_name, None) is None:
            self._localization_manager_cache[stream_name] = LocalizationManager(
                stream_name, self, throttle_hz
            )
        return self._localization_manager_cache[stream_name]

    @handle_agent_exceptions
    def get_agent_id(self):
        """
        Gets the Device ID for this device.

        :rtype: str
        """
        return self.get_agent_configuration().id

    @handle_agent_exceptions
    def get_agent_configuration(self):
        return self.agent_stub.GetAgentConfiguration(
            agent_pb2.GetAgentConfigurationRequest()
        ).configuration

    # Ingesting telemetry datapoints

    @handle_agent_exceptions
    def post_text(
        self,
        stream,  # type: str
        value,  # type: str
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        # type: (...) -> None
        """
        Post a text datapoint to a stream.

        :param stream: The name of the Formant stream to post the datapoint on
        :param value: The text datapoint value
        :param tags: Tags to include on the posted datapoint
        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
            Uses the current time by default
        :rtype: None

        .. highlight:: python
        .. code-block:: python

            from formant.sdk.agent.v1 import Client

            fclient = Client()
            fclient.post_text(
                "example.text",
                "Processed 9 items"
            )
        """

        datapoint = self.prepare_text(stream, value, tags, timestamp)
        self.post_data(datapoint)

    @handle_agent_exceptions
    def post_json(
        self,
        stream,  # type: str
        value,  # type: str
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        # type: (...) -> None
        """
        Post a JSON datapoint to a telemetry stream.

        :param stream: The name of the Formant stream to post the datapoint on
        :param value: The encoded JSON datapoint value
        :param tags: Tags to include on the posted datapoint
        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
            Uses the current time by default
        :rtype: None
        """

        datapoint = self.prepare_json(stream, value, tags, timestamp)
        self.post_data(datapoint)

    @handle_agent_exceptions
    def post_numeric(
        self,
        stream,  # type: str
        value,  # type: Union[float, int]
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        # type: (...) -> None
        """
        Post a numeric datapoint to a telemetry stream.

        :param stream: The name of the Formant stream to post the datapoint on
        :param value: The numeric datapoint value
        :param tags: Tags to include on the posted datapoint
        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
            Uses the current time by default
        :rtype: None
        """

        datapoint = self.prepare_numeric(stream, value, tags, timestamp)
        self.post_data(datapoint)

    @handle_agent_exceptions
    def post_numericset(
        self,
        stream,  # type: str
        numerics_dict,  # type: Dict[str, Tuple[Union[float, int], Optional[str]]]
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        # type: (...) -> None
        """
        Post a numeric set datapoint to a telemetry stream.
        Numeric sets are collections of related numeric datapoints.

        :param stream: The name of the Formant stream to post the datapoint on
        :param numerics_dict: The numeric set datapoint value,
            a dictionary mapping names to (numeric value, units) tuples.
        :param tags: Tags to include on the posted datapoint
        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
            Uses the current time by default
        :rtype: None

        .. highlight:: python
        .. code-block:: python

            from formant.sdk.agent.v1 import Client

            fclient = Client()
            fclient.post_numericset(
                "example.numericset2",
                {
                    "frequency": (998, "Hz"),
                    "usage": (30, "percent"),
                    "warp factor": (6.0, None),
                },
            )
        """

        datapoint = self.prepare_numericset(stream, numerics_dict, tags, timestamp)
        self.post_data(datapoint)

    @handle_agent_exceptions
    def post_image(
        self,
        stream,  # type: str
        value=None,  # type: Optional[bytes]
        url=None,  # type: Optional[str]
        content_type="image/jpg",
        # type: Literal["image/jpg", "image/png", "video/h264"]
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        # type: (...) -> None
        """
        Post an image datapoint to a telemetry stream.

        :param stream: The name of the Formant stream to post the datapoint on
        :param value: The datapoint value: raw bytes of an encoded image or frame
        :param url: The datapoint url: path to local file or valid remote URL for
            remote files
        :param content_type: The format of the encoded image or frame.
            Defaults to ``image/jpg``.
        :param tags: Tags to include on the posted datapoint
        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
            Uses the current time by default
        :rtype: None
        """
        datapoint = self.prepare_image(
            stream, value, url, content_type, tags, timestamp
        )
        self.post_data(datapoint)

    @handle_agent_exceptions
    def post_bitset(
        self,
        stream,  # type: str
        bitset_dict,  # type: Dict[str, bool]
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        # type: (...) -> None
        """
        Post a bitset datapoint to a telemetry stream.
        A bitset is a collection of related boolean states.

        :param stream: The name of the Formant stream to post the datapoint on
        :param bitset_dict: The datapoint value, a dictionary mapping names to booleans
        :param tags: Tags to include on the posted datapoint
        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
            Uses the current time by default
        :rtype: None

        .. highlight:: python
        .. code-block:: python

            from formant.sdk.agent.v1 import Client

            fclient = Client()
            fclient.post_bitset(
                "example.bitset",
                {
                    "standing": False,
                    "walking": False,
                    "sitting": True
                }
            )
        """

        datapoint = self.prepare_bitset(stream, bitset_dict, tags, timestamp)
        self.post_data(datapoint)

    @handle_agent_exceptions
    def post_geolocation(
        self,
        stream,  # type: str
        latitude,  # type: Union[float, int]
        longitude,  # type: Union[float, int]
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
        altitude=None,  # type: Union[float, int]
        orientation=None,  # type: Union[float, int]
    ):
        # type: (...) -> None
        """
        Post a geolocation datapoint to a telemetry stream.

        :param stream: The name of the Formant stream to post the datapoint on
        :param latitude: The datapoint value's latitude
        :param longitude: The datapoint value's longitude
        :param tags: Tags to include on the posted datapoint
        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
            Uses the current time by default
        :rtype: None
        """

        datapoint = self.prepare_geolocation(
            stream, latitude, longitude, tags, timestamp, altitude, orientation
        )
        self.post_data(datapoint)

    @handle_agent_exceptions
    def post_battery(
        self,
        stream,  # type: str
        percentage,  # type: Union[int, float]
        voltage=None,  # type: Optional[Union[int, float]]
        current=None,  # type: Optional[Union[int, float]]
        charge=None,  # type: Optional[Union[int, float]]
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        # type: (...) -> None
        """
        Post a battery datapoint to a telemetry stream. Only percentage is required.

        :param stream: The name of the Formant stream to post the datapoint on
        :param percentage: The battery charge percentage
        :param voltage: The battery voltage
        :param current: The battery current
        :param charge: The battery charge
        :param tags: Tags to include on the posted datapoint
        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
            Uses the current time by default
        :rtype: None
        """

        datapoint = self.prepare_battery(
            stream, percentage, voltage, current, charge, tags, timestamp
        )
        self.post_data(datapoint)

    @handle_agent_exceptions
    def post_file(
        self,
        stream,  # type: str
        url=None,  # type: str
        filename=None,  # type: Optional[str]
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        # type: (...) -> None
        """
        Post a file to a telemetry stream.

        :param stream: The name of the Formant stream to post the file on
        :param url: The file url: path to local file or valid remote URL for
            remote files
        :param filename: The file name: name displayed inside Formant module
        :param tags: Tags to include on the posted file
        :param timestamp: Unix timestamp in milliseconds for the posted file.
            Uses the current time by default
        :rtype: None

        .. highlight:: python
        .. code-block:: python

            from formant.sdk.agent.v1 import Client

            fclient = Client()
            fclient.post_file(
                "example.file",
                /home/user/Desktop/data/planets.csv,
                planets.csv,
            )
        """

        datapoint = self.prepare_file(stream, url, filename, tags, timestamp)
        self.post_data(datapoint)

    @handle_agent_exceptions
    @handle_grpc_exceptions
    def post_data(self, datapoint):
        # type: (...) -> None
        """
        :param datapoint:
        :rtype: None
        """
        self.agent_stub.PostData(datapoint)

    @handle_post_data_exceptions
    def post_data_multi(self, datapoints):
        # type: (...) -> None
        """
        :param datapoints:
        :rtype: None
        """
        request = agent_pb2.PostDataMultiRequest(datapoints=datapoints)
        self.agent_stub.PostDataMulti(request)

    def prepare_text(
        self,
        stream,  # type: str
        value,  # type: str
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        # type: (...) -> datapoint_pb2.Datapoint
        """
        Prepare a text datapoint without posting it.

        :param stream: The name of the Formant stream for the datapoint
        :param value: The text datapoint value
        :param tags: Tags for the datapoint
        :param timestamp: Unix timestamp in milliseconds for the datapoint.
            Uses the current time by default
        :rtype: datapoint_pb2.Datapoint
        """
        validate_type(value, [str])

        return datapoint_pb2.Datapoint(
            stream=stream,
            text=text_pb2.Text(value=value),
            tags=tags if tags else {},
            timestamp=timestamp if timestamp else current_timestamp(),
        )

    def prepare_json(
        self,
        stream,  # type: str
        value,  # type: str
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        # type (...) -> datapoint_pb2.Datapoint
        """
        Prepare a JSON datapoint without posting it.

        :param stream: (str) The name of the Formant stream for the datapoint
        :param value: (str) The encoded JSON datapoint value
        :param tags: (Optional[Dict[str, str]]) Tags for the datapoint
        :param timestamp: (Optional[int]) Unix timestamp in milliseconds for the datapoint.
            Uses the current time by default
        :rtype: datapoint_pb2.Datapoint
        """
        validate_type(value, [str])

        return datapoint_pb2.Datapoint(
            stream=stream,
            json=text_pb2.Json(value=value),
            timestamp=timestamp if timestamp else int(time.time() * 1000),
            tags=tags if tags else {},
        )

    def prepare_numeric(
        self,
        stream,  # type: str
        value,  # type: Union[float, int]
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        # type (...) -> datapoint_pb2.Datapoint
        """
        Prepare a numeric datapoint without posting it.

        :param stream: (str) The name of the Formant stream for the datapoint
        :param value: (Union[float, int]) The numeric datapoint value
        :param tags: (Optional[Dict[str, str]]) Tags for the datapoint
        :param timestamp: (Optional[int]) Unix timestamp in milliseconds for the datapoint.
            Uses the current time by default
        :rtype: datapoint_pb2.Datapoint
        """
        validate_type(value, [float, int])

        return datapoint_pb2.Datapoint(
            stream=stream,
            numeric=math_pb2.Numeric(value=value),
            tags=tags if tags else {},
            timestamp=timestamp if timestamp else current_timestamp(),
        )

    def prepare_numericset(
        self,
        stream,  # type: str
        numerics_dict,  # type: Dict[str, Tuple[Union[float, int], Optional[str]]]
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        # type (...) -> datapoint_pb2.Datapoint
        """
        Prepare a numeric set datapoint without posting it.

        :param stream: (str) The name of the Formant stream for the datapoint
        :param numerics_dict: (Dict[str, Tuple[Union[float, int], Optional[str]]]) The numeric set datapoint value,
            a dictionary mapping names to (numeric value, units) tuples.
        :param tags: (Optional[Dict[str, str]]) Tags for the datapoint
        :param timestamp: (Optional[int]) Unix timestamp in milliseconds for the datapoint.
            Uses the current time by default
        :rtype: The prepared numeric set datapoint
        :raises: ``TypeError: value v for key k in numericset must have length of 2``
        """
        validate_type(numerics_dict, [dict])

        numeric_set = math_pb2.NumericSet()
        for k, v in numerics_dict.items():
            validate_type(k, [str])
            validate_type(v, [tuple])
            if len(v) != 2:
                raise TypeError(
                    "value %s for key %s in numericset must have length of 2" % (v, k)
                )
            validate_type(v[0], [float, int])
            if v[1]:
                validate_type(v[1], [str])
            numeric_set.numerics.extend(
                [
                    math_pb2.NumericSetEntry(
                        value=v[0],
                        label=k,
                        unit=v[1] if v[1] else None,
                    )
                ]
            )

        return datapoint_pb2.Datapoint(
            stream=stream,
            numeric_set=numeric_set,
            tags=tags if tags else {},
            timestamp=timestamp if timestamp else current_timestamp(),
        )

    def prepare_image(
        self,
        stream,  # type: str
        value=None,  # type: Optional[bytes]
        url=None,  # type: Optional[str]
        content_type="image/jpg",
        # type: Literal["image/jpg", "image/png", "video/h264"]
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        # type (...) -> datapoint_pb2.Datapoint
        """
        Prepare an image datapoint without posting it.

        :param stream: (str) The name of the Formant stream for the datapoint
        :param value: (Optional[bytes]) The datapoint value: raw bytes of an encoded image or frame
        :param url: (Optional[str]) The datapoint url: path to a local file or valid remote URL for
            remote files
        :param content_type: (Literal["image/jpg", "image/png", "video/h264"]) The format of the encoded image or frame.
            Defaults to ``"image/jpg"``.
        :param tags: (Optional[Dict[str, str]]) Tags for the datapoint
        :param timestamp: (Optional[int]) Unix timestamp in milliseconds for the datapoint.
            Uses the current time by default
        :rtype: datapoint_pb2.Datapoint
        :raises: ``InvalidArgument: One of [url, value] must be used.``
        """

        if value is None and url is None:
            raise InvalidArgument("One of [url, value] must be used.")

        if value is not None and url is not None:
            raise InvalidArgument("Only one of [url, value] can be used.")

        if value is not None:
            validate_type(value, [bytes])

        if url is not None:
            validate_type(url, [str])

        validate_string_in_array(content_type, ALLOWED_IMAGE_AND_VIDEO_CONTENT_TYPES, "content_type")

        return datapoint_pb2.Datapoint(
            stream=stream,
            image=media_pb2.Image(raw=value, url=url, content_type=content_type),
            tags=tags if tags else {},
            timestamp=timestamp if timestamp else current_timestamp(),
        )

    def prepare_bitset(
        self,
        stream,  # type: str
        bitset_dict,  # type: Dict[str, bool]
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        # type(...) -> datapoint_pb2.Datapoint
        """
        Prepare a bitset datapoint without posting it.

        :param stream: (str) The name of the Formant stream for the datapoint
        :param bitset_dict: (Dict[str, bool]) The datapoint value, a dictionary mapping names to booleans
        :param tags: (Optional[Dict[str, str]]) Tags for the datapoint
        :param timestamp: (Optional[int]) Unix timestamp in milliseconds for the datapoint.
            Uses the current time by default
        :rtype: datapoint_pb2.Datapoint
        """
        validate_type(bitset_dict, [dict])

        bitset = math_pb2.Bitset()
        for k, v in bitset_dict.items():
            validate_type(k, [str])
            validate_type(v, [bool])
            bitset.bits.extend([math_pb2.Bit(key=k, value=v)])

        return datapoint_pb2.Datapoint(
            stream=stream,
            bitset=bitset,
            tags=tags if tags else {},
            timestamp=timestamp if timestamp else current_timestamp(),
        )

    def prepare_geolocation(
        self,
        stream,  # type: str
        latitude,  # type: Union[float, int]
        longitude,  # type: Union[float, int]
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
        altitude=None,  # type: Union[float, int]
        orientation=None,  # type: Union[float, int]
    ):
        """
        Prepare a geolocation datapoint without posting it.

        :param stream: (str) The name of the Formant stream for the datapoint
        :param latitude: (Union[float, int]) The datapoint value's latitude
        :param longitude: (Union[float, int]) The datapoint value's longitude
        :param tags: (Optional[Dict[str, str]]) Tags for the datapoint
        :param timestamp: (Optional[int]) Unix timestamp in milliseconds for the datapoint.
            Uses the current time by default
        :param altitude: (Union[float, int]) The altitude value (optional)
        :param orientation: (Union[float, int]) The orientation value (optional)
        :return: The prepared geolocation datapoint
        :rtype: datapoint_pb2.Datapoint
        """
        for _ in [latitude, longitude]:
            validate_type(_, [float, int])

        return datapoint_pb2.Datapoint(
            stream=stream,
            location=navigation_pb2.Location(
                latitude=latitude,
                longitude=longitude,
                altitude=altitude,
                orientation=orientation,
            ),
            tags=tags if tags else {},
            timestamp=timestamp if timestamp else current_timestamp(),
        )

    def prepare_battery(
        self,
        stream,  # type: str
        percentage,  # type: Union[int, float]
        voltage=None,  # type: Optional[Union[int, float]]
        current=None,  # type: Optional[Union[int, float]]
        charge=None,  # type: Optional[Union[int, float]]
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        """
        Prepare a battery datapoint without posting it.

        :param stream: (str) The name of the Formant stream for the datapoint
        :param percentage: (Union[int, float]) The battery charge percentage
        :param voltage: (Optional[Union[int, float]]) The battery voltage (optional)
        :param current: (Optional[Union[int, float]]) The battery current (optional)
        :param charge: (Optional[Union[int, float]]) The battery charge (optional)
        :param tags: (Optional[Dict[str, str]]) Tags for the datapoint
        :param timestamp: (Optional[int]) Unix timestamp in milliseconds for the datapoint.
            Uses the current time by default
        :return: The prepared battery datapoint
        :rtype: datapoint_pb2.Datapoint
        """
        for _ in [percentage, voltage, current, charge]:
            if _:
                validate_type(_, [float, int])

        battery = health_pb2.Battery()
        battery.percentage = percentage
        if voltage is not None:
            battery.voltage = voltage
        if current is not None:
            battery.current = current
        if charge is not None:
            battery.charge = charge

        return datapoint_pb2.Datapoint(
            stream=stream,
            battery=battery,
            tags=tags if tags else {},
            timestamp=timestamp if timestamp else current_timestamp(),
        )

    def prepare_file(
        self,
        stream,  # type: str
        url=None,  # type: str
        filename=None,  # type: Optional[str]
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
    ):
        """
        Prepare a file datapoint without posting it.

        :param stream: (str) The name of the Formant stream for the datapoint
        :param url: (str) The file url: path to a local file or valid remote URL for remote files
        :param filename: (Optional[str]) The file name: name displayed inside Formant module
        :param tags: (Optional[Dict[str, str]]) Tags for the datapoint
        :param timestamp: (Optional[int]) Unix timestamp in milliseconds for the datapoint.
            Uses the current time by default
        :return: The prepared file datapoint
        :rtype: datapoint_pb2.Datapoint
        """
        validate_type(filename, [str])
        validate_type(url, [str])

        return datapoint_pb2.Datapoint(
            stream=stream,
            file=file_pb2.File(url=url, filename=filename),
            tags=tags if tags else {},
            timestamp=timestamp if timestamp else current_timestamp(),
        )

    def register_telemetry_listener_callback(
        self,
        f,  # type: Callable[[datapoint_pb2.Datapoint], None]
        stream_filter=None,  # type: Optional[List[str]]
    ):
        # type : (...) -> None
        """
        Datapoints posted to the Formant agent whose "stream" value matches an element
        of the given stream filter will be streamed into the provided callback.
        If no stream filter is provided, datapoints from all streams
        will be received.

        :param f: A callback that will be called when a datapoint is posted
            to the Formant agent
        :param stream_filter: A list of stream names. The provided callback
            is only called for datapoints whose stream name is in this list
        """
        with self._lock:

            def create_stream():
                return self._get_telemetry_listener_stream(stream_filter=stream_filter)

            self._telemetry_listener_callback_streams[f] = CancellableStreamThread(
                f,
                create_stream,
                self.logger,
                ignore_unavailable=self.ignore_unavailable,
            )

    def unregister_telemetry_listener_callback(
        self, f  # type: Callable[[], None]
    ):
        # type : (...) -> None
        """
        Unregisters previously registered telemetry loopback callback.

        :param f: The telemetry loopback callback to be unregistered
        """
        with self._lock:
            stream = self._telemetry_listener_callback_streams.pop(f, None)
            if stream is not None:
                stream.cancel()

    # Ingesting transform frames

    @handle_agent_exceptions
    def set_base_frame_id(
        self, base_reference_frame  # type: str
    ):
        # type(...) -> None
        """
        Sets the base reference frame for tf tree ingestion.

        :param base_reference_frame: The base reference frame for the tf tree.
        :rtype: None
        """
        self.agent_stub.SetBaseFrameID(
            agent_pb2.SetBaseFrameIDRequest(id=base_reference_frame)
        )

    @handle_agent_exceptions
    def post_transform_frame(
        self,
        parent_frame,  # type: str
        child_frame,  # type: str
        tx,  # type: Union[int, float]
        ty,  # type: Union[int, float]
        tz,  # type: Union[int, float]
        rx,  # type: Union[int, float]
        ry,  # type: Union[int, float]
        rz,  # type: Union[int, float]
        rw,  # type: Union[int, float]
    ):
        # type: (...) -> None
        """
        Adds a transform frame, used to position datapoints in 3D space.

        :param parent_frame: The parent frame of the posted transform
        :param child_frame: The child frame of the posted transform
        :param tx: x-translation
        :param ty: y-translation
        :param tz: z-translation
        :param rx: x-rotation (quaternion)
        :param ry: y-rotation (quaternion)
        :param rz: z-rotation (quaternion)
        :param rw: w-rotation (quaternion)
        :rtype: None
        """

        for _ in [parent_frame, child_frame]:
            validate_type(_, [str])
        for _ in [tx, ty, tz, rx, ry, rz, rw]:
            validate_type(_, [float, int])

        frame = math_pb2.TransformFrame()
        frame.parent_frame = parent_frame
        frame.child_frame = child_frame
        frame.transform.translation.x = tx
        frame.transform.translation.y = ty
        frame.transform.translation.z = tz
        frame.transform.rotation.x = rx
        frame.transform.rotation.y = ry
        frame.transform.rotation.z = rz
        frame.transform.rotation.w = rw
        self.agent_stub.PostTransformFrame(frame)

    # Creating events

    @handle_agent_exceptions
    def create_event(
        self,
        message,  # type: str
        tags=None,  # type: Optional[Dict[str, str]]
        timestamp=None,  # type: Optional[int]
        end_timestamp=None,  # type: Optional[int]
        notify=False,  # type: bool
        severity=DEFAULT_SEVERITY_TYPE,
        # type: Literal["info", "warning", "critical", "error"]
    ):
        # type: (...) -> None
        """
        Creates and ingests an event.

        :param message: The text payload of the event
        :param tags: Tags to include on the event
        :param timestamp: Unix starting timestamp for the event.
            Uses the current time by default
        :param end_timestamp: Unix ending timestamp for the event.
            Must be greater than timestamp.
            If end_timestamp is supplied, the event will span a length of time
        :param notify: If True, the created event will trigger a Formant notification
        :param severity: The severity level of the event
        :rtype: None

        .. highlight:: python
        .. code-block:: python

            from formant.sdk.agent.v1 import Client

            fclient = Client()
            fclient.create_event(
                "Confinement beam to warp frequency 0.4e17 hz",
                tags={"Region": "North"},
                notify=True,
                severity="warning"
            )
        """

        validate_type(message, [str])
        validate_type(severity, [str])
        validate_string_in_array(severity, ALLOWED_SEVERITY_TYPES, "severity")

        severity_pb = set_severity_pb(severity)

        self.agent_stub.CreateEvent(
            agent_pb2.CreateEventRequest(
                event=event_pb2.Event(
                    message=message,
                    notification_enabled=notify,
                    tags=tags if tags else {},
                    severity=severity_pb,
                    timestamp=timestamp if timestamp else current_timestamp(),
                    end_timestamp=end_timestamp,
                )
            )
        )

    # Handling commands

    def get_command_request(
        self, command_filter=None  # type: Optional[List[str]]
    ):
        # type: (...) -> Optional[commands_pb2.CommandRequest]
        """
        If there is a command request in the agent's queue whose ``command`` value
        matches an element of the given command filter, takes and returns the
        command request.
        Otherwise, returns ``None`` if there are no matching command requests
        in the agent's queue.

        :param command_filter: A list of command names.
            This method only returns commands whose names are in this list.
        :rtype: CommandRequest, None
        """
        command_request_request = agent_pb2.GetCommandRequestRequest(
            command_filter=command_filter
        )
        try:
            command_request = self.agent_stub.GetCommandRequest(
                command_request_request
            ).request
            return command_request if command_request.command else None
        except grpc.RpcError:
            return None

    def send_command_response(
        self,
        request_id,  # type: str
        success,  # type: bool
        datapoint=None,  # type: Optional[datapoint_pb2.Datapoint]
    ):
        # type: (...) -> None
        """
        Sends a command response for an identified command request to Formant.
        Returns an error if there was a problem sending the command response.

        :param request_id: The ID of the command request to which this method responds
        :param success: Whether the command was successfully executed
        :param datapoint: A datapoint related to the command. Can attach a datapoint
            to a command response. E.g., if a command fails, can ingest a text datapoint
            with an error message related to the failure of the command.
        :rtype: None
        """
        response = commands_pb2.CommandResponse(
            request_id=request_id,
            success=success,
            datapoint=datapoint,
        )
        request = agent_pb2.SendCommandResponseRequest(response=response)
        self.agent_stub.SendCommandResponse(request)

    def register_command_request_callback(
        self,
        f,  # type: Callable[[commands_pb2.CommandRequest], None]
        command_filter=None,  # type: Optional[List[str]]
    ):
        # type: (...) -> None
        """
        Command requests issued to the agent whose ``command`` value matches an element
        of the given command filter will be streamed into the provided callback.
        If no command filter is provided, all command requests will be handled.

        :param f: A callback that will be executed on command requests
            as they are received by the Formant agent.
        :param command_filter: A list of command names. The provided callback
            is only executed on commands whose names are in this list
        :rtype: None
        """
        with self._lock:

            def create_stream():
                return self._get_command_request_stream(command_filter)

            self._command_request_callback_streams[f] = CancellableStreamThread(
                f,
                create_stream,
                self.logger,
                attribute="request",
                ignore_unavailable=self.ignore_unavailable,
            )

    def unregister_command_request_callback(
        self,
        f,  # type: Callable[[commands_pb2.CommandRequest], None]
    ):
        # type: (...) -> None
        """
        Unregisters previously registered command request callback.

        :param f: The command request callback to be unregistered
        :rtype: None
        """
        with self._lock:
            stream = self._command_request_callback_streams.pop(f, None)
            if stream is not None:
                stream.cancel()

    # Teleop

    def register_teleop_callback(
        self,
        f,  # type: Callable[[datapoint_pb2.ControlDatapoint], None]
        stream_filter=None,  # type: Optional[List[str]]
    ):
        # type: (...) -> None
        """
        Control datapoints received from teleop whose ``stream`` value matches an element
        of the given stream filter will be streamed into the provided callback.
        If no stream filter is provided, control datapoints from all streams
        will be received.

        :param f: A callback that will be executed on teleop control datapoints as they
            are received by the Formant agent
        :param stream_filter: A list of stream names. The provided callback
            is only exectued on control datapoints whose names are in this list
        :rtype: None
        """

        with self._lock:

            def create_stream():
                return self._get_teleop_stream(stream_filter)

            self._teleop_callback_streams[f] = CancellableStreamThread(
                f,
                create_stream,
                self.logger,
                attribute="control_datapoint",
                ignore_unavailable=self.ignore_unavailable,
            )

    def unregister_teleop_callback(
        self, f  # type: Callable[[datapoint_pb2.ControlDatapoint], None]
    ):
        # type: (...) -> None
        """
        Unregisters previously registered teleop callback.

        :param f: The teleop callback to be unregistered
        :rtype: None
        """
        with self._lock:
            stream = self._teleop_callback_streams.pop(f, None)
            if stream is not None:
                stream.cancel()

    @handle_agent_exceptions
    def get_teleop_info(self):
        # type: (...) -> agent_pb2.GetTeleopInfoResponse
        """
        Returns current information about teleop connection count.

        :rtype: GetTeleopInfoResponse
        """
        request = agent_pb2.GetTeleopInfoRequest()
        return self.agent_stub.GetTeleopInfo(request)

    def register_teleop_heartbeat_callback(
        self,
        f,  # type: Callable[[agent_pb2.GetTeleopHeartbeatStreamResponse], None]
    ):
        # type : (...) -> None
        """
        The provided callback will be called once each time a heartbeat
        is received over Formant teleop. Heartbeats are streamed from the
        operator machine at 20Hz on a UDP-like channel. This method can be used
        to quickly detect teleop disconnections.

        :param f: A callback that will be called when a heartbeat is received.
        :rtype: None
        """
        with self._lock:

            def create_stream():
                return self._get_teleop_heartbeat_stream()

            self._teleop_heartbeat_callback_streams[f] = CancellableStreamThread(
                f,
                create_stream,
                self.logger,
                ignore_unavailable=self.ignore_unavailable,
            )

    def unregister_teleop_heartbeat_callback(
        self, f  # type: Callable[[], None]
    ):
        # type : (...) -> None
        """
        Unregisters previously registered teleop heartbeat callback.

        :param f: The teleop heartbeat callback to be unregistered
        :rtype: None
        """
        with self._lock:
            stream = self._teleop_heartbeat_callback_streams.pop(f, None)
            if stream is not None:
                stream.cancel()

    # Custom data channel methods

    @handle_agent_exceptions
    def send_on_custom_data_channel(
        self,
        channel_name,  # type: str
        payload,  # type: bytes
    ):
        # type(...) -> None
        """
        Sends data on custom data channel.

        :param channel_name: (str) The name of the channel over which to send data
        :param payload: (bytes) The data payload to send.
        :rtype: None
        """
        self.agent_stub.SendOnCustomDataChannel(
            agent_pb2.SendOnCustomDataChannelRequest(
                channel_name=channel_name, payload=payload
            )
        )

    def register_custom_data_channel_message_callback(
        self,
        f,  # type: Callable[[agent_pb2.GetCustomDataChannelMessageStreamResponse], None] # noqa: E501
        channel_name_filter=None,  # type: Optional[List[str]]
    ):
        # type : (...) -> None
        """
        Registers a callback on data presence on the specified data channel.

        :param f: A callback that will be called with messages
            received on the specified custom data channel.
        :param channel_name_filter: An optional allow list of custom channel names
            for this callback.
        :rtype: None
        """
        with self._lock:

            def create_stream():
                return self._get_custom_data_channel_message_stream(
                    channel_name_filter=channel_name_filter
                )

            self._custom_data_channel_message_callback_streams[
                f
            ] = CancellableStreamThread(
                f,
                create_stream,
                self.logger,
                ignore_unavailable=self.ignore_unavailable,
            )

    def unregister_custom_data_channel_message_callback(
        self,
        f,  # type: Callable[[], None]
    ):
        # type : (...) -> None
        """
        Unregisters previously registered custom data channel callback. 

        :param f: The custom data channel message callback to be unregistered.
        :rtype: None
        """
        with self._lock:
            stream = self._custom_data_channel_message_callback_streams.pop(f, None)
            if stream is not None:
                stream.cancel()

    def custom_data_channel_request_handler(
        self, channel_name  # type: str
    ):
        # type : (...) -> Callable[Callable[agent_pb2.GetCustomDataChannelMessageStreamResponse, None], Callable[str, str]]: # noqa: E501
        """
        Registers a handler for requests sent by RequestDataChannel instances
        (part of the Formant toolkit).
        See: https://github.com/FormantIO/toolkit/tree/master/examples/request-response
        for an example.

        :param channel_name: The name of the custom data channel to listen on.
        :rtype: GetCustomDataChannelMessageStreamResponse

        .. highlight:: python
        .. code-block:: python

            from formant.sdk.agent.v1 import Client

            fclient = Client()

            @fclient.custom_data_channel_request_handler("my_channel")
            def handler(request_data):
                # Do something with request_data string
                print(json.loads(request_data))

                # Return any string response
                return json.dumps({"message": "Hello world!"})
        """

        def outer_wrapper(f):
            def inner_wrapper(message):
                try:
                    payload = json.loads(message.payload.decode("utf-8"))
                    request_id = payload["id"]
                    request_data = payload["data"]
                except (
                    UnicodeDecodeError,
                    json.decoder.JSONDecodeError,
                    KeyError,
                ) as e:
                    self.logger.error(
                        "Received invalid custom data channel request: "
                        + e.__class__.__name__
                        + ": "
                        + str(e)
                    )

                response_data = None
                try:
                    response = f(request_data)
                    # Place valid responses on the "data" key
                    response_data = json.dumps(
                        {"id": request_id, "data": response}
                    ).encode("utf-8")
                except Exception as e:
                    error_message = (
                        "An error occurred while handling custom data channel request: "
                        + e.__class__.__name__
                        + ": "
                        + str(e)
                    )
                    # Place error responses on the "error" key
                    response_data = json.dumps(
                        {"id": request_id, "error": error_message}
                    ).encode("utf-8")
                self.send_on_custom_data_channel(channel_name, response_data)

            # When the handler function is decorated,
            # we automatically register it as a callback
            self.register_custom_data_channel_message_callback(
                inner_wrapper, [channel_name]
            )

        return outer_wrapper

    def custom_data_channel_binary_request_handler(
        self,
        channel_name,  # type: bytes
        new_thread=False,  # type: bool
    ):
        # type : (...) -> Callable[Callable[agent_pb2.GetCustomDataChannelMessageStreamResponse, None], Callable[str, str]]: # noqa: E501
        """
        :param channel_name: The name of the custom data channel to listen on.

        .. highlight:: python
        .. code-block:: python

            from formant.sdk.agent.v1 import Client

            fclient = Client()

            @fclient.custom_data_channel_request_handler("my_channel")
            def handler(request_data):
                # Do something with request_data bytes
                print(request_data.decode("utf-8"))

                # Return any bytes response
                return b"Hello."
        """

        SUCCESS_RESPONSE_BYTE = b"\00"
        ERROR_RESPONSE_BYTE = b"\01"

        def outer_wrapper(f):
            def inner_wrapper(message):
                try:
                    request_id = message.payload[0:16]
                    request_data = message.payload[16:]
                except TypeError as e:
                    self.logger.error(
                        "Received invalid custom data channel request: "
                        + e.__class__.__name__
                        + ": "
                        + str(e)
                    )

                def call_and_respond():
                    response_data = None
                    try:
                        response = f(request_data)
                        response_data = request_id + SUCCESS_RESPONSE_BYTE + response
                    except Exception as e:
                        error_message = (
                            "An error occurred while handling custom data channel request: "
                            + e.__class__.__name__
                            + ": "
                            + str(e)
                        )
                        response_data = (
                            request_id
                            + ERROR_RESPONSE_BYTE
                            + error_message.encode("utf-8")
                        )

                    self.send_on_custom_data_channel(channel_name, response_data)

                if new_thread:
                    self._thread_pool.apply_async(call_and_respond)
                else:
                    call_and_respond()

            # When the handler function is decorated,
            # we automatically register it as a callback
            self.register_custom_data_channel_message_callback(
                inner_wrapper, [channel_name]
            )

        return outer_wrapper

    # Reading application configuration

    def get_app_config(self, key, *args):
        # type: (str, Any) -> Optional[str]
        """
        Returns the value for the given key that was set in
        Formant application configuration for this device,
        or returns the given default value.

        :param key: The application configuration key
        :param args: (One additional argument)
            The default value to return if the key is not found.
        :raises: ``TypeError: Function takes at most two args: (key: str, default: Any)``
        """
        if len(args) > 1:
            raise TypeError("function takes at most two args: (key: str, default: Any)")
        default = args[0] if len(args) == 1 else None

        # update if there's a connection to the agent API
        self._update_application_configuration()
        return self._app_config.get(key, default)

    @handle_agent_exceptions
    def get_config_blob_data(self):
        # type: (...) -> str
        """
        Returns the blob data defined in the device configuration.

        :rtype: str
        """
        request = agent_pb2.GetConfigBlobDataRequest()
        return self.agent_stub.GetConfigBlobData(request).blob_data.data

    @handle_agent_exceptions
    def get_buffer_metadata(self):
        # type: (...) -> agent_pb2.GetBufferMetadata
        """
        Returns the current WebRTC buffer statistics.

        :rtype: agent_pb2.GetBufferMetadata
        """
        request = agent_pb2.GetBufferMetadataRequest()
        return self.agent_stub.GetBufferMetadata(request)

    def register_config_update_callback(self, f):
        # type: (Callable) -> None
        """
        Adds a function to the list of callbacks that are executed by the client
        when this device receives updated configuration from Formant.

        :param f: The configuration update callback to be registered.
        :rtype: None
        """
        with self._lock:
            self._config_update_callbacks.append(f)

    def unregister_config_update_callback(self, f):
        # type: (Callable) -> None
        """
        Removes a function from the list of callbacks that are executed by the client
        when this device receives updated configuration from Formant.

        :param f: The configuration update callback to be unregistered.
        :rtype: None
        """
        with self._lock:
            self._config_update_callbacks.remove(f)

    # Internal methods


    def _setup_agent_communication(self, subscribe_only=False):
        if subscribe_only:
            self.channel.subscribe(self._handle_connectivity_change, try_to_connect=True)
        else:
            self.channel = grpc.insecure_channel(self._agent_url)
            self.channel.subscribe(self._handle_connectivity_change, try_to_connect=True)
            self.agent_stub = agent_pb2_grpc.AgentStub(self.channel)

    def _handle_connectivity_change(self, connectivity):
        """Handle changes to gRPC channel connectivity."""

        self.logger.info("Agent communication status: %s" % connectivity.name)

        if connectivity == grpc.ChannelConnectivity.READY and not self._connected:
            self.logger.info("Agent communication READY. Updating application configuration.")
            self._connected = True
            self._update_application_configuration()
            self._run_update_config_callbacks()
        elif connectivity == grpc.ChannelConnectivity.IDLE and self._connected:
            # If the channel is back to idle we have to re-subscribe before we can use it
            # this will transition to READY if the connection is re-established
            self._connected = False
            self.logger.info("Agent communication IDLE. Re-subscribing...")
            self._setup_agent_communication(subscribe_only=True)
        elif connectivity == grpc.ChannelConnectivity.SHUTDOWN:
            # In the case of shutdown, re-establish the connection from scratch
            self.logger.info("Agent communication lost. Re-establishing...")
            self._setup_agent_communication()

    @handle_agent_exceptions
    def _update_application_configuration(self):
        request = agent_pb2.GetApplicationConfigurationRequest()
        response = self.agent_stub.GetApplicationConfiguration(request)
        self._app_config = response.configuration.configuration_map

    def _run_update_config_callbacks(self):
        [f() for f in self._config_update_callbacks]

    def _get_command_request_stream(self, command_filter=None):
        return self.agent_stub.GetCommandRequestStream(
            agent_pb2.GetCommandRequestStreamRequest(command_filter=command_filter)
        )

    def _get_teleop_stream(self, stream_filter=None):
        return self.agent_stub.GetTeleopControlDataStream(
            agent_pb2.GetTeleopControlDataStreamRequest(stream_filter=stream_filter)
        )

    def _get_teleop_heartbeat_stream(self):
        return self.agent_stub.GetTeleopHeartbeatStream(
            agent_pb2.GetTeleopHeartbeatStreamRequest()
        )

    def _get_telemetry_listener_stream(self, stream_filter=None):
        return self.agent_stub.GetTelemetryListenerStream(
            agent_pb2.GetTelemetryListenerStreamRequest(stream_filter=stream_filter)
        )

    def _get_custom_data_channel_message_stream(self, channel_name_filter=None):
        return self.agent_stub.GetCustomDataChannelMessageStream(
            agent_pb2.GetCustomDataChannelMessageStreamRequest(
                channel_name_filter=channel_name_filter
            )
        )

    @handle_agent_exceptions
    def call_cloud(
        self,
        endpoint,  # type: str
        method,  # type: str
        body,  # type: str
        headers,  # type: Dict[str, str]
        require_formant_auth,  # type: bool
        buffer_call,  # type: bool
        is_retryable,  # type: bool
        retryable_status_codes=[]  # type: List[int]
    ):
    
        """
        Allows the user to call an endpoint of the Formant Admin API
        authenticated by the Formant agent instead of user credentials.

        API calls which allow ``device`` authentication can buffer and retry
        calls. For more information, see the following documentation:

        `Use the Formant agent to authenticate API calls <https://docs.formant.io/reference/use-the-formant-agent-to-authenticate-api-calls>`__
        
        `Buffering and retrying API calls <https://docs.formant.io/reference/buffering-and-retrying-api-calls>`__

        .. note::

            If buffering is enabled, you will not get a return value from this function.

        :param endpoint: Full URL of the endpoint to call (can be found at https://docs.formant.io/reference).
        :type endpoint: str
        :param method: The HTTP method to use (e.g., ``"POST"``, ``"PUT"``, ``"GET"``, ``"PATCH"``, ``"DELETE"``).
        :type method: str
        :param headers: Set the content type of your payload.
        :type headers: Dict[str, str]
        :param body: Payload of the request (parameters found at https://docs.formant.io/reference).
        :type body: str
        :param require_formant_auth: Whether or not to use device authentication.
                    If ``True``, ``authorization`` header is added automatically.
        :type require_formant_auth: bool
        :param buffer_call: Whether or not to buffer the call.
                    If ``True``, the call is buffered and will be retried if necessary.
                    If ``True``, ``call_cloud()`` returns ``None``.
        :type buffer_call: bool
        :param is_retryable: (``buffer_call=True`` only) Whether to retry the call if it fails.
        :type is_retryable: bool
        :param retryable_status_codes: (``buffer_call=True`` only) The status codes to retry on.
                    A value of ``[-1]`` will retry on all ``5xx`` codes EXCEPT FOR the following: ``[500, 501, 502, 505, 507, 508, 510, 511]``.
        :type retryable_status_codes: List[int]
        :rtype: agent_pb2.PostGenericAPIUnbufferedRequestResponse

        .. code-block:: python

            from formant.sdk.agent.v1 import Client
            import json

            fclient = Client()

            payload = {
                "query": "acme",
                "count": 10
            }

            response = fclient.call_cloud(
                endpoint="https://api.formant.io/v1/admin/devices/query",
                method="POST", 
                headers={
                    "Content-Type": "application/json"
                },
                body=json.dumps(payload),
                require_formant_auth=True,
                buffer_call=False,
                is_retryable=False,
                retryable_status_codes=[]
            )
            
            # You get a response with ``statusCode`` and ``responseBody`` 
            # when ``buffer_call == False``.
            print(response.statusCode)
            print(response.responseBody)
        """
        validate_type(endpoint, [str])
        validate_type(method, [str])
        validate_type(body, [str])
        validate_type(headers, [dict])
        validate_type(require_formant_auth, [bool])
        validate_type(is_retryable, [bool])
        validate_type(retryable_status_codes, [list])

        request = datapoint_pb2.GenericAPIDatapoint()
        request.Method = method
        request.Endpoint = endpoint
        for key, value in headers.items():
            request.Headers[key] = value

        request.Body = body
        request.RequireFormantAuth = require_formant_auth
        request.IsRetryable = is_retryable
        for code in retryable_status_codes:
            request.RetryableStatusCodes.append(code)
        if buffer_call:
            return self.agent_stub.PostGenericAPIRequest(request)

        return self.agent_stub.PostGenericAPIUnbufferedRequest(request)

    @handle_agent_exceptions
    def create_selection_intervention_request(
        self,
        title,  # type: str
        instruction,  # type: str
        options,  # type: List[str]
        hint,  # type: int
        url=None,  # type: str
        content_type=DEFAULT_IMAGE_CONTENT_TYPE, # type: Literal["image/jpg", "image/png"]
        timestamp=None,  # type: Optional[int]
        severity=DEFAULT_SEVERITY_TYPE, # type: Literal["info", "warning", "critical", "error"]
    ):
        # type : (...) -> intervention_pb2.InterventionRequest
        """
        Creates an intervention request based on type ``selection``.
        Takes an image url, options and an integer with an optional
        addition of instructions, and title.

        :param title: (str) The name of the intervention
        :param instruction: (str) The instructions detailing how to resolve the intervention
        :param options: (List[str]) The list with options to select from
        :param hint: (int) The index of the suspected correct answer
        :param url: (str) The path to local file or valid remote URL for remote files
        :param content_type: (Literal["image/jpg", "image/png"]) The format of the encoded image or frame.
            Defaults to ``"image/jpg"``
        :param timestamp: (Optional[int]) Unix timestamp in milliseconds for the posted datapoint.
            Uses the current time by default
        :param severity: (Literal["info", "warning", "critical", "error"]) The severity level of the event
        :rtype: InterventionRequest

        .. highlight:: python
        .. code-block:: python

            from formant.sdk.agent.v1 import Client

            fclient = Client()
            fclient.create_selection_intervention_request(
                "Which fruit is best?",
                "Select the best grape",
                ["fruit_1", "fruit_2", "fruit_3"],
                hint=1,
                url=/home/my_user/data/test-image.jpeg
                severity=critical
            )
        """

        validate_type(title, [str])
        validate_type(instruction, [str])
        validate_type(url, [str])
        validate_type(options, [List])
        validate_type(hint, [int])
        validate_type(severity, [str])
        validate_type(content_type, [str])
        validate_string_in_array(
            content_type, ALLOWED_IMAGE_CONTENT_TYPES, "content_type"
        )
        validate_string_in_array(severity, ALLOWED_SEVERITY_TYPES, "severity")
        for option in options:
            validate_type(option, [str])

        severity_pb = set_severity_pb(severity)

        request = intervention_pb2.InterventionRequest()
        request.timestamp = timestamp if timestamp else current_timestamp()
        request.severity = severity_pb
        request.selection_request.title = title
        request.selection_request.hint = hint
        request.selection_request.instruction = instruction
        request.selection_request.image.content_type = content_type
        request.selection_request.image.url = url
        request.selection_request.options.extend(options)
        return self.agent_stub.CreateInterventionRequest(request)

    @handle_agent_exceptions
    def create_labeling_intervention_request(
        self,
        title,  # type: str
        instruction,  # type: str
        labels,  # type: Dict[str, str]
        hint=None,  # type: Optional[List[intervention_pb2.LabeledPolygon]]
        url=None,  # type: str
        content_type=DEFAULT_IMAGE_CONTENT_TYPE, # type: Literal["image/jpg", "image/png"]
        timestamp=None,  # type: Optional[int]
        severity=DEFAULT_SEVERITY_TYPE,
        # type: Literal["info", "warning", "critical", "error"]
    ):
        # type : (...) -> intervention_pb2.InterventionRequest

        """
        Creates an intervention request based on type "labeling".

        :param title: (str) The name of the intervention
        :param instruction: (str) The instructions detailing how to resolve the intervention
        :param labels: (Dict[str, str]) An Array of labels
        :param hint: (Optional[List[intervention_pb2.LabeledPolygon]]) An array of label polygons, X and Y coordinates with a label
        :param url: (str) The path to local file or valid remote URL for remote files
        :param content_type: (Literal["image/jpg", "image/png"]) The format of the encoded image or frame.
            Defaults to ``"image/jpg"``.
        :param timestamp: (Optional[int]) Unix timestamp in milliseconds for the posted datapoint.
            Uses the current time by default
        :param severity: (Literal["info", "warning", "critical", "error"]) The severity level of the event
        :rtype: intervention_pb2.InterventionRequest


        Each ``label`` in ``labels`` defined as::

                Label = {
                        value = string;
                        string display_name = string;
                        }
        
        Hint is an array of "LabeledPolygon", defined as::

                hint = {
                    List of vertex,
                    List of labels
                }

        where each vertex is defined as::

                vertex = {
                    x = float,
                    y = float
                }
        """
        validate_type(title, [str])
        validate_type(instruction, [str])
        validate_type(url, [str])
        validate_type(content_type, [str])
        validate_type(severity, [str])
        validate_string_in_array(severity, ALLOWED_SEVERITY_TYPES, "severity")
        validate_string_in_array(
            content_type, ALLOWED_IMAGE_CONTENT_TYPES, "content_type"
        )
        validate_type(labels, [dict])
        if hint is not None:
            for label in hint:
                validate_type(label, [intervention_pb2.LabeledPolygon])
        for k, v in labels.items():
            validate_type(k, [str])
            validate_type(v, [str])

        severity_pb = set_severity_pb(severity)

        request = intervention_pb2.InterventionRequest()
        request.timestamp = timestamp if timestamp else current_timestamp()
        request.severity = severity_pb
        request.labeling_request.title = title
        request.labeling_request.instruction = instruction
        if hint is not None:
            request.labeling_request.hint.extend(hint)
        request.labeling_request.image.content_type = content_type
        request.labeling_request.image.url = url
        for key, value in labels.items():
            label = intervention_pb2.Label()
            label.value = key
            label.display_name = value
            request.labeling_request.labels.append(label)
        return self.agent_stub.CreateInterventionRequest(request)

    @handle_agent_exceptions
    def get_intervention_response(
        self,
        request_id,  # type: str
    ):
        # type : (...) -> agent_pb2.GetInterventionResponse

        """
        Receives request ID, and returns a response.
        
        :param request_id: (str) The ID of the intervention request
            to which this method responds
        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
            Uses the current time by default.
        :rtype: agent_pb2.GetInterventionResponse


        .. highlight:: python
        .. code-block:: python

            from formant.sdk.agent.v1 import Client

            fclient = Client()
            request = fclient.create_selection_intervention_request(
                title="",
                instruction="instruction",
                options=["option1", "option2", "option3"],
                hint=0,
                url="/home/formantuser/Downloads/image.png",
            )
            fclient.handle_intervention_response(request.id)
        """
        validate_type(request_id, [str])
        response_request = agent_pb2.GetInterventionResponseRequest()
        response_request.request_id = request_id
        return self.agent_stub.GetInterventionResponse(response_request)



from formant.sdk.cloud.v2.src.resources.resources import Resources
from formant.sdk.cloud.v2.formant_query_api_client.api.online_devices import online_devices_controller_online

class OnlineDevices(Resources):

    def online(self):
        """See devices online currently"""
        client = self._get_client()
        response = online_devices_controller_online.sync_detailed(client=client)
        return response

    async def online_async(self):
        """See devices online currently"""
        client = self._get_client()
        response = await online_devices_controller_online.asyncio_detailed(client=client)
        return response


from formant.sdk.cloud.v2.formant_admin_api_client.models import CommandParameterMeta, CommandListResponse, CommandTemplateParameterMeta, CommandTemplate, CommandParameter, CommandDeliveryEventType, CommandDeliveryEvent, CommandDeliveryEventTags, CommandResponseTags, CommandRequest, CommandResponseStreamType, FileInfo, CommandQuery, CommandDeliveryEventStreamType, CommandProgress, Filter, CommandRequestListResponse, ScopeFilter, CommandDeliveryEventMetadata, CommandTemplateListResponse, ScopeFilterTypesItem, CommandTemplateTags, CommandStreamType, Command, CommandResponse, CommandTags
from formant.sdk.cloud.v2.formant_admin_api_client.api.command import command_controller_post, command_controller_query
from formant.sdk.cloud.v2.formant_admin_api_client.types import *


from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="EventTriggerCommand")


@attr.s(auto_attribs=True)
class EventTriggerCommand:
    """
    Attributes:
        command_template_id (str):
        parameter_value (Union[Unset, str]):
    """

    command_template_id: str
    parameter_value: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        command_template_id = self.command_template_id
        parameter_value = self.parameter_value

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "commandTemplateId": command_template_id,
            }
        )
        if parameter_value is not UNSET:
            field_dict["parameterValue"] = parameter_value

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        command_template_id = d.pop("commandTemplateId")

        parameter_value = d.pop("parameterValue", UNSET)

        event_trigger_command = cls(
            command_template_id=command_template_id,
            parameter_value=parameter_value,
        )

        event_trigger_command.additional_properties = d
        return event_trigger_command

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import Any, Dict, List, Type, TypeVar, Union, cast

import attr
from dateutil.parser import isoparse

from ..models.event_seek_query_direction import EventSeekQueryDirection
from ..models.event_seek_query_event_types_item import EventSeekQueryEventTypesItem
from ..models.event_seek_query_severities_item import EventSeekQuerySeveritiesItem
from ..models.event_seek_query_types_item import EventSeekQueryTypesItem
from ..types import UNSET, Unset

T = TypeVar("T", bound="EventSeekQuery")


@attr.s(auto_attribs=True)
class EventSeekQuery:
    """
    Attributes:
        direction (EventSeekQueryDirection):
        from_ (datetime.datetime):
        tags (Union[Unset, Any]):
        device_ids (Union[Unset, List[str]]):
        names (Union[Unset, List[str]]):
        types (Union[Unset, List[EventSeekQueryTypesItem]]):
        not_tags (Union[Unset, Any]):
        not_names (Union[Unset, List[str]]):
        agent_ids (Union[Unset, List[str]]):
        start (Union[Unset, datetime.datetime]):
        end (Union[Unset, datetime.datetime]):
        id (Union[Unset, str]):
        viewed (Union[Unset, bool]):
        keyword (Union[Unset, str]):
        message (Union[Unset, str]):
        event_types (Union[Unset, List[EventSeekQueryEventTypesItem]]):
        notification_enabled (Union[Unset, bool]):
        user_ids (Union[Unset, List[str]]):
        annotation_template_ids (Union[Unset, List[str]]):
        disable_null_matches (Union[Unset, bool]):
        severities (Union[Unset, List[EventSeekQuerySeveritiesItem]]):
        sets_device_color (Union[Unset, bool]):
    """

    direction: EventSeekQueryDirection
    from_: datetime.datetime
    tags: Union[Unset, Any] = UNSET
    device_ids: Union[Unset, List[str]] = UNSET
    names: Union[Unset, List[str]] = UNSET
    types: Union[Unset, List[EventSeekQueryTypesItem]] = UNSET
    not_tags: Union[Unset, Any] = UNSET
    not_names: Union[Unset, List[str]] = UNSET
    agent_ids: Union[Unset, List[str]] = UNSET
    start: Union[Unset, datetime.datetime] = UNSET
    end: Union[Unset, datetime.datetime] = UNSET
    id: Union[Unset, str] = UNSET
    viewed: Union[Unset, bool] = UNSET
    keyword: Union[Unset, str] = UNSET
    message: Union[Unset, str] = UNSET
    event_types: Union[Unset, List[EventSeekQueryEventTypesItem]] = UNSET
    notification_enabled: Union[Unset, bool] = UNSET
    user_ids: Union[Unset, List[str]] = UNSET
    annotation_template_ids: Union[Unset, List[str]] = UNSET
    disable_null_matches: Union[Unset, bool] = UNSET
    severities: Union[Unset, List[EventSeekQuerySeveritiesItem]] = UNSET
    sets_device_color: Union[Unset, bool] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        direction = self.direction.value

        from_ = self.from_.isoformat()

        tags = self.tags
        device_ids: Union[Unset, List[str]] = UNSET
        if not isinstance(self.device_ids, Unset):
            device_ids = self.device_ids

        names: Union[Unset, List[str]] = UNSET
        if not isinstance(self.names, Unset):
            names = self.names

        types: Union[Unset, List[str]] = UNSET
        if not isinstance(self.types, Unset):
            types = []
            for types_item_data in self.types:
                types_item = types_item_data.value

                types.append(types_item)

        not_tags = self.not_tags
        not_names: Union[Unset, List[str]] = UNSET
        if not isinstance(self.not_names, Unset):
            not_names = self.not_names

        agent_ids: Union[Unset, List[str]] = UNSET
        if not isinstance(self.agent_ids, Unset):
            agent_ids = self.agent_ids

        start: Union[Unset, str] = UNSET
        if not isinstance(self.start, Unset):
            start = self.start.isoformat()

        end: Union[Unset, str] = UNSET
        if not isinstance(self.end, Unset):
            end = self.end.isoformat()

        id = self.id
        viewed = self.viewed
        keyword = self.keyword
        message = self.message
        event_types: Union[Unset, List[str]] = UNSET
        if not isinstance(self.event_types, Unset):
            event_types = []
            for event_types_item_data in self.event_types:
                event_types_item = event_types_item_data.value

                event_types.append(event_types_item)

        notification_enabled = self.notification_enabled
        user_ids: Union[Unset, List[str]] = UNSET
        if not isinstance(self.user_ids, Unset):
            user_ids = self.user_ids

        annotation_template_ids: Union[Unset, List[str]] = UNSET
        if not isinstance(self.annotation_template_ids, Unset):
            annotation_template_ids = self.annotation_template_ids

        disable_null_matches = self.disable_null_matches
        severities: Union[Unset, List[str]] = UNSET
        if not isinstance(self.severities, Unset):
            severities = []
            for severities_item_data in self.severities:
                severities_item = severities_item_data.value

                severities.append(severities_item)

        sets_device_color = self.sets_device_color

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "direction": direction,
                "from": from_,
            }
        )
        if tags is not UNSET:
            field_dict["tags"] = tags
        if device_ids is not UNSET:
            field_dict["deviceIds"] = device_ids
        if names is not UNSET:
            field_dict["names"] = names
        if types is not UNSET:
            field_dict["types"] = types
        if not_tags is not UNSET:
            field_dict["notTags"] = not_tags
        if not_names is not UNSET:
            field_dict["notNames"] = not_names
        if agent_ids is not UNSET:
            field_dict["agentIds"] = agent_ids
        if start is not UNSET:
            field_dict["start"] = start
        if end is not UNSET:
            field_dict["end"] = end
        if id is not UNSET:
            field_dict["id"] = id
        if viewed is not UNSET:
            field_dict["viewed"] = viewed
        if keyword is not UNSET:
            field_dict["keyword"] = keyword
        if message is not UNSET:
            field_dict["message"] = message
        if event_types is not UNSET:
            field_dict["eventTypes"] = event_types
        if notification_enabled is not UNSET:
            field_dict["notificationEnabled"] = notification_enabled
        if user_ids is not UNSET:
            field_dict["userIds"] = user_ids
        if annotation_template_ids is not UNSET:
            field_dict["annotationTemplateIds"] = annotation_template_ids
        if disable_null_matches is not UNSET:
            field_dict["disableNullMatches"] = disable_null_matches
        if severities is not UNSET:
            field_dict["severities"] = severities
        if sets_device_color is not UNSET:
            field_dict["setsDeviceColor"] = sets_device_color

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        direction = EventSeekQueryDirection(d.pop("direction"))

        from_ = isoparse(d.pop("from"))

        tags = d.pop("tags", UNSET)

        device_ids = cast(List[str], d.pop("deviceIds", UNSET))

        names = cast(List[str], d.pop("names", UNSET))

        types = []
        _types = d.pop("types", UNSET)
        for types_item_data in _types or []:
            types_item = EventSeekQueryTypesItem(types_item_data)

            types.append(types_item)

        not_tags = d.pop("notTags", UNSET)

        not_names = cast(List[str], d.pop("notNames", UNSET))

        agent_ids = cast(List[str], d.pop("agentIds", UNSET))

        _start = d.pop("start", UNSET)
        start: Union[Unset, datetime.datetime]
        if isinstance(_start, Unset):
            start = UNSET
        else:
            start = isoparse(_start)

        _end = d.pop("end", UNSET)
        end: Union[Unset, datetime.datetime]
        if isinstance(_end, Unset):
            end = UNSET
        else:
            end = isoparse(_end)

        id = d.pop("id", UNSET)

        viewed = d.pop("viewed", UNSET)

        keyword = d.pop("keyword", UNSET)

        message = d.pop("message", UNSET)

        event_types = []
        _event_types = d.pop("eventTypes", UNSET)
        for event_types_item_data in _event_types or []:
            event_types_item = EventSeekQueryEventTypesItem(event_types_item_data)

            event_types.append(event_types_item)

        notification_enabled = d.pop("notificationEnabled", UNSET)

        user_ids = cast(List[str], d.pop("userIds", UNSET))

        annotation_template_ids = cast(List[str], d.pop("annotationTemplateIds", UNSET))

        disable_null_matches = d.pop("disableNullMatches", UNSET)

        severities = []
        _severities = d.pop("severities", UNSET)
        for severities_item_data in _severities or []:
            severities_item = EventSeekQuerySeveritiesItem(severities_item_data)

            severities.append(severities_item)

        sets_device_color = d.pop("setsDeviceColor", UNSET)

        event_seek_query = cls(
            direction=direction,
            from_=from_,
            tags=tags,
            device_ids=device_ids,
            names=names,
            types=types,
            not_tags=not_tags,
            not_names=not_names,
            agent_ids=agent_ids,
            start=start,
            end=end,
            id=id,
            viewed=viewed,
            keyword=keyword,
            message=message,
            event_types=event_types,
            notification_enabled=notification_enabled,
            user_ids=user_ids,
            annotation_template_ids=annotation_template_ids,
            disable_null_matches=disable_null_matches,
            severities=severities,
            sets_device_color=sets_device_color,
        )

        event_seek_query.additional_properties = d
        return event_seek_query

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.device_stream_file_tail_configuration_file_format import DeviceStreamFileTailConfigurationFileFormat
from ..models.device_stream_file_tail_configuration_type import DeviceStreamFileTailConfigurationType
from ..types import UNSET, Unset

T = TypeVar("T", bound="DeviceStreamFileTailConfiguration")


@attr.s(auto_attribs=True)
class DeviceStreamFileTailConfiguration:
    """
    Attributes:
        type (DeviceStreamFileTailConfigurationType):
        filename (str):
        file_format (Union[Unset, DeviceStreamFileTailConfigurationFileFormat]):
        time_key (Union[Unset, str]):
        time_format (Union[Unset, str]):
        regex (Union[Unset, str]):
    """

    type: DeviceStreamFileTailConfigurationType
    filename: str
    file_format: Union[Unset, DeviceStreamFileTailConfigurationFileFormat] = UNSET
    time_key: Union[Unset, str] = UNSET
    time_format: Union[Unset, str] = UNSET
    regex: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        type = self.type.value

        filename = self.filename
        file_format: Union[Unset, str] = UNSET
        if not isinstance(self.file_format, Unset):
            file_format = self.file_format.value

        time_key = self.time_key
        time_format = self.time_format
        regex = self.regex

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type,
                "filename": filename,
            }
        )
        if file_format is not UNSET:
            field_dict["fileFormat"] = file_format
        if time_key is not UNSET:
            field_dict["timeKey"] = time_key
        if time_format is not UNSET:
            field_dict["timeFormat"] = time_format
        if regex is not UNSET:
            field_dict["regex"] = regex

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        type = DeviceStreamFileTailConfigurationType(d.pop("type"))

        filename = d.pop("filename")

        _file_format = d.pop("fileFormat", UNSET)
        file_format: Union[Unset, DeviceStreamFileTailConfigurationFileFormat]
        if isinstance(_file_format, Unset):
            file_format = UNSET
        else:
            file_format = DeviceStreamFileTailConfigurationFileFormat(_file_format)

        time_key = d.pop("timeKey", UNSET)

        time_format = d.pop("timeFormat", UNSET)

        regex = d.pop("regex", UNSET)

        device_stream_file_tail_configuration = cls(
            type=type,
            filename=filename,
            file_format=file_format,
            time_key=time_key,
            time_format=time_format,
            regex=regex,
        )

        device_stream_file_tail_configuration.additional_properties = d
        return device_stream_file_tail_configuration

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

import attr

if TYPE_CHECKING:
    from ..models.ros_topic import RosTopic


T = TypeVar("T", bound="DeviceRosState")


@attr.s(auto_attribs=True)
class DeviceRosState:
    """
    Attributes:
        topics (List['RosTopic']):
    """

    topics: List["RosTopic"]
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        topics = []
        for topics_item_data in self.topics:
            topics_item = topics_item_data.to_dict()

            topics.append(topics_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "topics": topics,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.ros_topic import RosTopic

        d = src_dict.copy()
        topics = []
        _topics = d.pop("topics")
        for topics_item_data in _topics:
            topics_item = RosTopic.from_dict(topics_item_data)

            topics.append(topics_item)

        device_ros_state = cls(
            topics=topics,
        )

        device_ros_state.additional_properties = d
        return device_ros_state

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import Any, Dict, List, Type, TypeVar, Union, cast

import attr
from dateutil.parser import isoparse

from ..models.sql_query_aggregate_level import SqlQueryAggregateLevel
from ..models.sql_query_aggregate_type import SqlQueryAggregateType
from ..models.sql_query_types_item import SqlQueryTypesItem
from ..types import UNSET, Unset

T = TypeVar("T", bound="SqlQuery")


@attr.s(auto_attribs=True)
class SqlQuery:
    """
    Attributes:
        sql_query (Union[Unset, str]):
        parameters (Union[Unset, List[str]]):
        start (Union[Unset, datetime.datetime]):
        end (Union[Unset, datetime.datetime]):
        limit (Union[Unset, float]):
        table_columns (Union[Unset, List[Any]]):
        aggregate_level (Union[Unset, SqlQueryAggregateLevel]):
        aggregate_type (Union[Unset, SqlQueryAggregateType]):
        order_by_column (Union[Unset, str]):
        order_by_descending (Union[Unset, bool]):
        group_by_device (Union[Unset, bool]):
        time_dimension (Union[Unset, bool]):
        type (Union[Unset, str]):
        fill_gaps (Union[Unset, bool]):
        unit (Union[Unset, str]):
        timezone (Union[Unset, str]):
        account_id (Union[Unset, str]):
        tags (Union[Unset, Any]):
        device_ids (Union[Unset, List[str]]):
        names (Union[Unset, List[str]]):
        types (Union[Unset, List[SqlQueryTypesItem]]):
        not_tags (Union[Unset, Any]):
        not_names (Union[Unset, List[str]]):
        agent_ids (Union[Unset, List[str]]):
    """

    sql_query: Union[Unset, str] = UNSET
    parameters: Union[Unset, List[str]] = UNSET
    start: Union[Unset, datetime.datetime] = UNSET
    end: Union[Unset, datetime.datetime] = UNSET
    limit: Union[Unset, float] = UNSET
    table_columns: Union[Unset, List[Any]] = UNSET
    aggregate_level: Union[Unset, SqlQueryAggregateLevel] = UNSET
    aggregate_type: Union[Unset, SqlQueryAggregateType] = UNSET
    order_by_column: Union[Unset, str] = UNSET
    order_by_descending: Union[Unset, bool] = UNSET
    group_by_device: Union[Unset, bool] = UNSET
    time_dimension: Union[Unset, bool] = UNSET
    type: Union[Unset, str] = UNSET
    fill_gaps: Union[Unset, bool] = UNSET
    unit: Union[Unset, str] = UNSET
    timezone: Union[Unset, str] = UNSET
    account_id: Union[Unset, str] = UNSET
    tags: Union[Unset, Any] = UNSET
    device_ids: Union[Unset, List[str]] = UNSET
    names: Union[Unset, List[str]] = UNSET
    types: Union[Unset, List[SqlQueryTypesItem]] = UNSET
    not_tags: Union[Unset, Any] = UNSET
    not_names: Union[Unset, List[str]] = UNSET
    agent_ids: Union[Unset, List[str]] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        sql_query = self.sql_query
        parameters: Union[Unset, List[str]] = UNSET
        if not isinstance(self.parameters, Unset):
            parameters = self.parameters

        start: Union[Unset, str] = UNSET
        if not isinstance(self.start, Unset):
            start = self.start.isoformat()

        end: Union[Unset, str] = UNSET
        if not isinstance(self.end, Unset):
            end = self.end.isoformat()

        limit = self.limit
        table_columns: Union[Unset, List[Any]] = UNSET
        if not isinstance(self.table_columns, Unset):
            table_columns = self.table_columns

        aggregate_level: Union[Unset, str] = UNSET
        if not isinstance(self.aggregate_level, Unset):
            aggregate_level = self.aggregate_level.value

        aggregate_type: Union[Unset, str] = UNSET
        if not isinstance(self.aggregate_type, Unset):
            aggregate_type = self.aggregate_type.value

        order_by_column = self.order_by_column
        order_by_descending = self.order_by_descending
        group_by_device = self.group_by_device
        time_dimension = self.time_dimension
        type = self.type
        fill_gaps = self.fill_gaps
        unit = self.unit
        timezone = self.timezone
        account_id = self.account_id
        tags = self.tags
        device_ids: Union[Unset, List[str]] = UNSET
        if not isinstance(self.device_ids, Unset):
            device_ids = self.device_ids

        names: Union[Unset, List[str]] = UNSET
        if not isinstance(self.names, Unset):
            names = self.names

        types: Union[Unset, List[str]] = UNSET
        if not isinstance(self.types, Unset):
            types = []
            for types_item_data in self.types:
                types_item = types_item_data.value

                types.append(types_item)

        not_tags = self.not_tags
        not_names: Union[Unset, List[str]] = UNSET
        if not isinstance(self.not_names, Unset):
            not_names = self.not_names

        agent_ids: Union[Unset, List[str]] = UNSET
        if not isinstance(self.agent_ids, Unset):
            agent_ids = self.agent_ids

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if sql_query is not UNSET:
            field_dict["sqlQuery"] = sql_query
        if parameters is not UNSET:
            field_dict["parameters"] = parameters
        if start is not UNSET:
            field_dict["start"] = start
        if end is not UNSET:
            field_dict["end"] = end
        if limit is not UNSET:
            field_dict["limit"] = limit
        if table_columns is not UNSET:
            field_dict["tableColumns"] = table_columns
        if aggregate_level is not UNSET:
            field_dict["aggregateLevel"] = aggregate_level
        if aggregate_type is not UNSET:
            field_dict["aggregateType"] = aggregate_type
        if order_by_column is not UNSET:
            field_dict["orderByColumn"] = order_by_column
        if order_by_descending is not UNSET:
            field_dict["orderByDescending"] = order_by_descending
        if group_by_device is not UNSET:
            field_dict["groupByDevice"] = group_by_device
        if time_dimension is not UNSET:
            field_dict["timeDimension"] = time_dimension
        if type is not UNSET:
            field_dict["type"] = type
        if fill_gaps is not UNSET:
            field_dict["fillGaps"] = fill_gaps
        if unit is not UNSET:
            field_dict["unit"] = unit
        if timezone is not UNSET:
            field_dict["timezone"] = timezone
        if account_id is not UNSET:
            field_dict["accountId"] = account_id
        if tags is not UNSET:
            field_dict["tags"] = tags
        if device_ids is not UNSET:
            field_dict["deviceIds"] = device_ids
        if names is not UNSET:
            field_dict["names"] = names
        if types is not UNSET:
            field_dict["types"] = types
        if not_tags is not UNSET:
            field_dict["notTags"] = not_tags
        if not_names is not UNSET:
            field_dict["notNames"] = not_names
        if agent_ids is not UNSET:
            field_dict["agentIds"] = agent_ids

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        sql_query = d.pop("sqlQuery", UNSET)

        parameters = cast(List[str], d.pop("parameters", UNSET))

        _start = d.pop("start", UNSET)
        start: Union[Unset, datetime.datetime]
        if isinstance(_start, Unset):
            start = UNSET
        else:
            start = isoparse(_start)

        _end = d.pop("end", UNSET)
        end: Union[Unset, datetime.datetime]
        if isinstance(_end, Unset):
            end = UNSET
        else:
            end = isoparse(_end)

        limit = d.pop("limit", UNSET)

        table_columns = cast(List[Any], d.pop("tableColumns", UNSET))

        _aggregate_level = d.pop("aggregateLevel", UNSET)
        aggregate_level: Union[Unset, SqlQueryAggregateLevel]
        if isinstance(_aggregate_level, Unset):
            aggregate_level = UNSET
        else:
            aggregate_level = SqlQueryAggregateLevel(_aggregate_level)

        _aggregate_type = d.pop("aggregateType", UNSET)
        aggregate_type: Union[Unset, SqlQueryAggregateType]
        if isinstance(_aggregate_type, Unset):
            aggregate_type = UNSET
        else:
            aggregate_type = SqlQueryAggregateType(_aggregate_type)

        order_by_column = d.pop("orderByColumn", UNSET)

        order_by_descending = d.pop("orderByDescending", UNSET)

        group_by_device = d.pop("groupByDevice", UNSET)

        time_dimension = d.pop("timeDimension", UNSET)

        type = d.pop("type", UNSET)

        fill_gaps = d.pop("fillGaps", UNSET)

        unit = d.pop("unit", UNSET)

        timezone = d.pop("timezone", UNSET)

        account_id = d.pop("accountId", UNSET)

        tags = d.pop("tags", UNSET)

        device_ids = cast(List[str], d.pop("deviceIds", UNSET))

        names = cast(List[str], d.pop("names", UNSET))

        types = []
        _types = d.pop("types", UNSET)
        for types_item_data in _types or []:
            types_item = SqlQueryTypesItem(types_item_data)

            types.append(types_item)

        not_tags = d.pop("notTags", UNSET)

        not_names = cast(List[str], d.pop("notNames", UNSET))

        agent_ids = cast(List[str], d.pop("agentIds", UNSET))

        sql_query = cls(
            sql_query=sql_query,
            parameters=parameters,
            start=start,
            end=end,
            limit=limit,
            table_columns=table_columns,
            aggregate_level=aggregate_level,
            aggregate_type=aggregate_type,
            order_by_column=order_by_column,
            order_by_descending=order_by_descending,
            group_by_device=group_by_device,
            time_dimension=time_dimension,
            type=type,
            fill_gaps=fill_gaps,
            unit=unit,
            timezone=timezone,
            account_id=account_id,
            tags=tags,
            device_ids=device_ids,
            names=names,
            types=types,
            not_tags=not_tags,
            not_names=not_names,
            agent_ids=agent_ids,
        )

        sql_query.additional_properties = d
        return sql_query

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

import attr
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.partial_event_trigger_group_sms_tags import PartialEventTriggerGroupSmsTags
    from ..models.partial_event_trigger_group_tags import PartialEventTriggerGroupTags
    from ..models.scope_filter import ScopeFilter


T = TypeVar("T", bound="PartialEventTriggerGroup")


@attr.s(auto_attribs=True)
class PartialEventTriggerGroup:
    """
    Attributes:
        organization_id (Union[Unset, str]):
        enabled (Union[Unset, bool]):
        device_scope (Union[Unset, None, ScopeFilter]):
        sms_tags (Union[Unset, PartialEventTriggerGroupSmsTags]):
        phone_numbers (Union[Unset, List[str]]): User's phone numbers (e.g., +14155550100).
        id (Union[Unset, str]):
        created_at (Union[Unset, datetime.datetime]):
        updated_at (Union[Unset, datetime.datetime]):
        tags (Union[Unset, PartialEventTriggerGroupTags]):
    """

    organization_id: Union[Unset, str] = UNSET
    enabled: Union[Unset, bool] = UNSET
    device_scope: Union[Unset, None, "ScopeFilter"] = UNSET
    sms_tags: Union[Unset, "PartialEventTriggerGroupSmsTags"] = UNSET
    phone_numbers: Union[Unset, List[str]] = UNSET
    id: Union[Unset, str] = UNSET
    created_at: Union[Unset, datetime.datetime] = UNSET
    updated_at: Union[Unset, datetime.datetime] = UNSET
    tags: Union[Unset, "PartialEventTriggerGroupTags"] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        organization_id = self.organization_id
        enabled = self.enabled
        device_scope: Union[Unset, None, Dict[str, Any]] = UNSET
        if not isinstance(self.device_scope, Unset):
            device_scope = self.device_scope.to_dict() if self.device_scope else None

        sms_tags: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.sms_tags, Unset):
            sms_tags = self.sms_tags.to_dict()

        phone_numbers: Union[Unset, List[str]] = UNSET
        if not isinstance(self.phone_numbers, Unset):
            phone_numbers = self.phone_numbers

        id = self.id
        created_at: Union[Unset, str] = UNSET
        if not isinstance(self.created_at, Unset):
            created_at = self.created_at.isoformat()

        updated_at: Union[Unset, str] = UNSET
        if not isinstance(self.updated_at, Unset):
            updated_at = self.updated_at.isoformat()

        tags: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if organization_id is not UNSET:
            field_dict["organizationId"] = organization_id
        if enabled is not UNSET:
            field_dict["enabled"] = enabled
        if device_scope is not UNSET:
            field_dict["deviceScope"] = device_scope
        if sms_tags is not UNSET:
            field_dict["smsTags"] = sms_tags
        if phone_numbers is not UNSET:
            field_dict["phoneNumbers"] = phone_numbers
        if id is not UNSET:
            field_dict["id"] = id
        if created_at is not UNSET:
            field_dict["createdAt"] = created_at
        if updated_at is not UNSET:
            field_dict["updatedAt"] = updated_at
        if tags is not UNSET:
            field_dict["tags"] = tags

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.partial_event_trigger_group_sms_tags import PartialEventTriggerGroupSmsTags
        from ..models.partial_event_trigger_group_tags import PartialEventTriggerGroupTags
        from ..models.scope_filter import ScopeFilter

        d = src_dict.copy()
        organization_id = d.pop("organizationId", UNSET)

        enabled = d.pop("enabled", UNSET)

        _device_scope = d.pop("deviceScope", UNSET)
        device_scope: Union[Unset, None, ScopeFilter]
        if _device_scope is None:
            device_scope = None
        elif isinstance(_device_scope, Unset):
            device_scope = UNSET
        else:
            device_scope = ScopeFilter.from_dict(_device_scope)

        _sms_tags = d.pop("smsTags", UNSET)
        sms_tags: Union[Unset, PartialEventTriggerGroupSmsTags]
        if isinstance(_sms_tags, Unset):
            sms_tags = UNSET
        else:
            sms_tags = PartialEventTriggerGroupSmsTags.from_dict(_sms_tags)

        phone_numbers = cast(List[str], d.pop("phoneNumbers", UNSET))

        id = d.pop("id", UNSET)

        _created_at = d.pop("createdAt", UNSET)
        created_at: Union[Unset, datetime.datetime]
        if isinstance(_created_at, Unset):
            created_at = UNSET
        else:
            created_at = isoparse(_created_at)

        _updated_at = d.pop("updatedAt", UNSET)
        updated_at: Union[Unset, datetime.datetime]
        if isinstance(_updated_at, Unset):
            updated_at = UNSET
        else:
            updated_at = isoparse(_updated_at)

        _tags = d.pop("tags", UNSET)
        tags: Union[Unset, PartialEventTriggerGroupTags]
        if isinstance(_tags, Unset):
            tags = UNSET
        else:
            tags = PartialEventTriggerGroupTags.from_dict(_tags)

        partial_event_trigger_group = cls(
            organization_id=organization_id,
            enabled=enabled,
            device_scope=device_scope,
            sms_tags=sms_tags,
            phone_numbers=phone_numbers,
            id=id,
            created_at=created_at,
            updated_at=updated_at,
            tags=tags,
        )

        partial_event_trigger_group.additional_properties = d
        return partial_event_trigger_group

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
